from dataclasses import dataclass

from nwn.gff._impl import FieldKind


class Byte(int):
    FIELD_KIND = FieldKind.BYTE

    def __new__(cls, value):
        if not 0 <= value <= 255:
            raise ValueError(f"BYTE value out of bounds: {value}")
        return super().__new__(cls, value)


class Char(int):
    FIELD_KIND = FieldKind.CHAR

    def __new__(cls, value):
        if not -128 <= value <= 127:
            raise ValueError(f"CHAR value out of bounds: {value}")
        return super().__new__(cls, value)


class Word(int):
    FIELD_KIND = FieldKind.WORD

    def __new__(cls, value):
        if not 0 <= value <= 65535:
            raise ValueError(f"WORD value out of bounds: {value}")
        return super().__new__(cls, value)


class Short(int):
    FIELD_KIND = FieldKind.SHORT

    def __new__(cls, value):
        if not -32768 <= value <= 32767:
            raise ValueError(f"SHORT value out of bounds: {value}")
        return super().__new__(cls, value)


class Dword(int):
    FIELD_KIND = FieldKind.DWORD

    def __new__(cls, value):
        if not 0 <= value <= 4294967295:
            raise ValueError(f"DWORD value out of bounds: {value}")
        return super().__new__(cls, value)


class Int(int):
    FIELD_KIND = FieldKind.INT

    def __new__(cls, value):
        if not -2147483648 <= value <= 2147483647:
            raise ValueError(f"INT value out of bounds: {value}")
        return super().__new__(cls, value)


class Dword64(int):
    FIELD_KIND = FieldKind.DWORD64

    def __new__(cls, value):
        if not 0 <= value <= 18446744073709551615:
            raise ValueError(f"DWORD64 value out of bounds: {value}")
        return super().__new__(cls, value)


class Int64(int):
    FIELD_KIND = FieldKind.INT64

    def __new__(cls, value):
        if not -9223372036854775808 <= value <= 9223372036854775807:
            raise ValueError(f"INT64 value out of bounds: {value}")
        return super().__new__(cls, value)


class Float(float):
    FIELD_KIND = FieldKind.FLOAT


class Double(float):
    FIELD_KIND = FieldKind.DOUBLE


class CExoString(str):
    FIELD_KIND = FieldKind.CEXOSTRING


class ResRef(str):
    FIELD_KIND = FieldKind.RESREF

    def __new__(cls, value):
        if len(value) > 16:
            raise ValueError(f"ResRef value too long: {value}")
        return super().__new__(cls, value)


@dataclass
class CExoLocString:
    """Represents a localized string in the NWN engine."""

    FIELD_KIND = FieldKind.CEXOLOCSTRING

    strref: int
    entries: dict[int, str]


class VOID(bytes):
    FIELD_KIND = FieldKind.VOID

    def __new__(cls, value):
        return super().__new__(cls, value)


class Struct(dict):
    """GFF Structs are just python dicts with .attr access and some metadata."""

    FIELD_KIND = FieldKind.STRUCT

    def __init__(self, struct_id, **kwargs):
        super().__init__(**kwargs)
        object.__setattr__(self, "_struct_id", struct_id)

    @property
    def struct_id(self):
        """The struct ID of the struct."""
        return object.__getattribute__(self, "_struct_id")

    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError as exc:
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{item}'"
            ) from exc

    def __setattr__(self, name, value):
        self[name] = value

    def __hash__(self):
        return hash(self.struct_id)


class List(list[Struct]):
    """
    GFF Lists are just python lists of Structs. They carry no metadata.

    This class exists as a convenience for type checking and
    future extensibility.
    """

    FIELD_KIND = FieldKind.LIST

    def __init__(self, value=None):
        if value is None:
            value = []
        super().__init__(value)



"""
Read and write TLK (Talk Table) files.

All strings are transparently converted to and from the NWN encoding.
"""

import struct

from typing import NamedTuple, BinaryIO
from nwn._shared import Language, get_nwn_encoding


class TlkEntry(NamedTuple):
    """
    A single entry in a TLK file including the text, sound resref, and sound length.
    Only used when reading TLK files with sound data.
    """

    text: str
    sound_resref: str = ""
    sound_length: float = 0.0

    def __str__(self):
        return self.text


def read(
    file: BinaryIO, include_sound_data=False, max_entries=0x7FFFF
) -> tuple[list, Language]:
    """
    Reads a TLK file fully into memory and returns a list of entries.

    Args:
        file: A binary file object containing the TLK file.
        include_sound_data: If True, entries are returned as TlkEntry objects.
            Otherwise, the resulting dict will contain only the text.
        max_entries: The maximum number of entries to read from the TLK file.
            This is a sanity check to avoid allocating excess memory when
            reading untrusted or corrupted data.

    Returns:
        list: An object containing the language and entries from the TLK file.
        Language: The language the TLK file is configured as.

    Raises:
        ValueError: If the file does not contain valid TLK data.
    """

    magic = file.read(4)
    if magic != b"TLK ":
        raise ValueError("Invalid TLK magic")
    version = file.read(4)
    if version != b"V3.0":
        raise ValueError("Invalid TLK version")
    language = struct.unpack("<I", file.read(4))[0]
    language = Language(language)
    entry_count = struct.unpack("<I", file.read(4))[0]
    entries_offset = struct.unpack("<I", file.read(4))[0]

    if entry_count > max_entries:
        raise ValueError(f"Too many entries in TLK file: {entry_count} > {max_entries}")

    entries = []
    for i in range(entry_count):
        file.seek(20 + 40 * i)

        (
            _,  # flags (text_present=0x1, sound_present=0x2,length_present=0x4)
            sound_resref,
            _,  # volume variance: unused as per spec
            _,  # pitch variance: unused as per spec
            offset_to_string,
            string_sz,
            sound_length,
        ) = struct.unpack("<I16sIIIIf", file.read(40))

        sound_resref = sound_resref.decode("ascii").strip("\x00\xc0")

        file.seek(entries_offset + offset_to_string)
        text = file.read(string_sz).decode(get_nwn_encoding())

        if include_sound_data:
            entries.append(TlkEntry(text, sound_resref, sound_length))
        else:
            entries.append(text)

    return (entries, language)


def write(file: BinaryIO, entries: list, language: Language):
    """
    Writes a Tlk object to a binary file.

    Args:
        file: A binary file object to write the TLK data to.
        entries: A list containing the entries to write, in order.
            Entries can be either strings or TlkEntry objects.
        language: The language of the TLK file to write.

    Raises:
        ValueError: If the TLK object contains invalid data.
    """

    file.write(
        struct.pack(
            "<4s4sIII",
            b"TLK ",
            b"V3.0",
            language.value,
            len(entries),
            20 + len(entries) * 40,
        )
    )

    str_data = bytearray()

    string_offset = 0
    for idx, entry in enumerate(entries):
        flags = 0
        text_len = 0

        if isinstance(entry, str):
            entry = TlkEntry(entry)

        if entry.text:
            flags |= 0x1
            text_len = len(entry.text.encode(get_nwn_encoding()))
        if entry.sound_resref:
            flags |= 0x2
        if entry.sound_length:
            flags |= 0x4

        if len(entry.sound_resref) > 16:
            raise ValueError(f"Sound resref at {idx} is too long")

        file.write(
            struct.pack(
                "<I16sIIIIf",
                flags,
                entry.sound_resref.ljust(16, "\x00").encode("ascii"),
                0,  # volume variance: unused as per spec
                0,  # pitch variance: unused as per spec
                string_offset,
                text_len,
                entry.sound_length,
            )
        )
        string_offset += text_len
        str_data += entry.text.encode(get_nwn_encoding())

    file.write(str_data)
